/**
 * Architektur & Implementierung von DBMS, Aufgabenblatt 7
 *
 * Bitmap Indexing, Query Processing and WAH Compression
 *
 * File        : PlainBitVector.h
 */

#ifndef PLAINBITVECTOR_H
#define PLAINBITVECTOR_H

#include <stdint.h>
#include <bitset>
#include <vector>

class PlainBitVector;

/***** Global bit operations on PlainBitVectors *****/
PlainBitVector operator& (const PlainBitVector& lhs, const PlainBitVector& rhs);
PlainBitVector operator| (const PlainBitVector& lhs, const PlainBitVector& rhs);


/***** PlainBitVector class represents uncompressed bit vectors *****/
class PlainBitVector {
  private:
    uint32_t mSize;
    uint32_t mNumWords;
    std::vector < std::bitset <32> > data;
    uint32_t itIdx;

  public:
    PlainBitVector(uint32_t size) : mSize(size)
      {
          mNumWords = size / 32 + 1;
	  data.resize(mNumWords);
      }

    /** number of bits */
    uint32_t size() const {
        return mSize;
    }

    /** number of 32-bit words required to store the bit vector */
    uint32_t getNumWords() const {
        return this->mNumWords;
    }

    friend PlainBitVector operator& (const PlainBitVector& lhs, const PlainBitVector& rhs);
    friend PlainBitVector operator| (const PlainBitVector& lhs, const PlainBitVector& rhs);

    /**
     * \defgroup BitOperations Bit operation methods over the bit vector.
     * @{ 
     */
    /* In-place AND operator. Assumption vector lengths will be same. */
    const PlainBitVector & operator&= (const PlainBitVector& rhs) {
        
        for(uint32_t i=0; i<mNumWords; i++) {
            data[i] &= rhs.data[i];
        }
        
        return *this;
    }

    /* In-place OR operator. Assumption vector lengths will be same. */
    const PlainBitVector & operator|= (const PlainBitVector& rhs) {
        for(uint32_t i=0; i<mNumWords; i++) {
            data[i] |= rhs.data[i];
        }
        return *this;
    }

    /** Count set bits in a bit vector */
    uint32_t count() const {
        uint32_t count = 0;
        for(uint32_t i=0; i<mNumWords; i++) {
            count += data[i].count();
        }
        return count;
    }
    /**@}*/

    /** 
     * When building the bitvector, this method is called to inform about the
     * next set-bit's (bit with value 1) index.
     *
     * Subsequent calls of this method must have a greater idx than the previous
     * call. 
     * 
     * @param idx Index of the next set-bit in the bitvector.
     */
    void nextSetBitIndex(uint32_t idx) {
        data[idx/32][idx%32]=1;
    }

    /**
     * \defgroup BitIteration Iteration methods over the bits of the bit vector.
     * @{ 
     */
    void begin() {
        itIdx=0;
    }

    /** has more bits? */
    bool hasNext() {
        return itIdx < mSize;
    }

    /** return next bit, 0:false, 1:true */
    bool getNext() {
        bool result = data[itIdx/32][itIdx%32];
        itIdx++;
        return result;
    }
    /**@}*/
};

#endif /* PLAINBITVECTOR_H */
