/**
 * Architektur & Implementierung von DBMS, Aufgabenblatt 7
 *
 * Bitmap Indexing, Query Processing and WAH Compression
 *
 * File        : queries.cpp
 */

#include <string.h>
#include <stdio.h>
#include <iostream>

#include "queries.h"
#include "BitmapIndex.h"
#include "FieldType.h"

/******************************************************************************
 *
 * QUERY IMPLEMENTATIONS WITHOUT BITMAP INDEXES
 *
 *****************************************************************************/

/** 
 * Query 1: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber = 3 OR linenumber = 5;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @return returns number of matching rows.
 */
int32_t
query1(uint32_t numrows, lineitem_t  *inputtable)
{
    /* Example implementation: */
    uint32_t count = 0;
    register uint32_t idx = 0;

    for(idx = 0; idx < numrows; ++idx){

        if(inputtable[idx].linenumber == 3 ||  inputtable[idx].linenumber == 5){

            count ++;

        }
    }

    return count;
}

/** 
 * Query 2: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber = 5      AND
 *           shipmode   = 'RAIL' AND
 *           quantity   = 10;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @return returns number of matching rows.
 */
int32_t
query2(uint32_t numrows, lineitem_t  *inputtable)
{
    uint32_t count = 0;
    register uint32_t idx = 0;

    for(idx = 0; idx < numrows; ++idx){

        if(inputtable[idx].linenumber == 5 && strcmp(inputtable[idx].shipmode, "RAIL") == 0 && inputtable[idx].quantity == 10){

            count ++;

        }
    }

    return count;
}

/** 
 * Query 3: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber IN (2, 4) AND
 *           shipmode   IN ('SHIP', 'TRUCK');
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @return returns number of matching rows.
 */
int32_t
query3(uint32_t numrows, lineitem_t  *inputtable)
{
    uint32_t count = 0;
    register uint32_t idx = 0;

    for(idx = 0; idx < numrows; ++idx){

        if((inputtable[idx].linenumber == 2 || inputtable[idx].linenumber == 4) && (strcmp(inputtable[idx].shipmode, "SHIP") == 0 || strcmp(inputtable[idx].shipmode, "TRUCK") == 0)){

            count ++;

        }
    }

    return count;
}

/** 
 * Query 4: Execute the following TPCH query:
 *    SELECT sum(extendedprice * (1.0-discount/100.0))
 *    FROM   lineitem
 *    WHERE  returnflag = 'R' AND
 *           shipmode IN ('REG AIR', 'AIR') AND
 *           quantity BETWEEN 5 AND 10;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @return returns the sum value.
 */
double
query4(uint32_t numrows, lineitem_t  *inputtable)
{
    double sum = 0.0;
    register uint32_t idx = 0;

    for(idx = 0; idx < numrows; ++idx){

        if  (inputtable[idx].returnflag == 'R' 
         && (inputtable[idx].quantity >= 5 && inputtable[idx].quantity <= 10) 
         && (strcmp(inputtable[idx].shipmode, "REG AIR") == 0 || strcmp(inputtable[idx].shipmode, "AIR") == 0)){

            sum += inputtable[idx].extendedprice * (1.0 - inputtable[idx].discount / 100.0);

        }
    }

    return sum;

}

/******************************************************************************
 *
 * QUERY IMPLEMENTATIONS WITH BITMAP INDEXES
 *
 ******************************************************************************/

/** 
 * Query 1: Execute the following TPCH query using bitmap-index on linenumber
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber = 3 OR linenumber = 5;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @param bitmap_idx 
 * 
 * @return returns number of matching rows.
 */
int32_t
query1_with_index(uint32_t numrows, lineitem_t  *inputtable, 
                  BitmapIndex<int32_t> & bitmap_idx_linenumber)
{
    // get all rows with linenumber = 3
    const BitVector & vec_ln3 = bitmap_idx_linenumber.getBitVector(3);
    // get all rows with linenumber = 5
    const BitVector & vec_ln5 = bitmap_idx_linenumber.getBitVector(5);
    // OR bitvectors
    BitVector res = vec_ln3 | vec_ln5;
    // return count
    return res.count();
}

/** 
 * Query 2: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber = 5      AND
 *           shipmode   = 'RAIL' AND
 *           quantity   = 10;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @param bitmap_idx_linenumber bitmap index on linenumber
 * @param bitmap_idx_shipmode bitmap index on shipmode
 * @param bitmap_idx_quantity bitmap index on quantity
 * 
 * @return returns number of matching rows.
 */
int32_t
query2_with_index(uint32_t numrows, lineitem_t  *inputtable,
                  BitmapIndex<int32_t>  & bitmap_idx_linenumber,
                  BitmapIndex<char[11]> & bitmap_idx_shipmode,
                  BitmapIndex<int64_t>  & bitmap_idx_quantity)
{
    const BitVector & vec_ln5 = bitmap_idx_linenumber.getBitVector(5);
    const BitVector & vec_q10 = bitmap_idx_quantity.getBitVector(10);
    char rail[11] = "RAIL";
    const BitVector & vec_rail = bitmap_idx_shipmode.getBitVector(rail);
    
    return (vec_ln5 & vec_rail & vec_q10).count();
/*
    BitVector vec_ln5 = bitmap_idx_linenumber.getBitVector(5);
    const BitVector & vec_q10 = bitmap_idx_quantity.getBitVector(10);
    char rail[11] = "RAIL";
    const BitVector & vec_rail = bitmap_idx_shipmode.getBitVector(rail);
 
    vec_ln5 &= vec_rail;
    vec_ln5 &= vec_q10;    
    return vec_ln5.count();
*/    

}

/** 
 * Query 3: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber IN (2, 4) AND
 *           shipmode   IN ('SHIP', 'TRUCK');
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @param bitmap_idx_linenumber bitmap index on linenumber
 * @param bitmap_idx_shipmode bitmap index on shipmode
 * 
 * @return returns number of matching rows.
 */
int32_t
query3_with_index(uint32_t numrows, lineitem_t  *inputtable,
                  BitmapIndex<int32_t>  & bitmap_idx_linenumber,
                  BitmapIndex<char[11]> & bitmap_idx_shipmode)
{
    char ship[11] = "SHIP";
    const BitVector & vec_ship = bitmap_idx_shipmode.getBitVector(ship);
    char truck[11] = "TRUCK";
    const BitVector & vec_truck = bitmap_idx_shipmode.getBitVector(truck);
    const BitVector & vec_ln2 = bitmap_idx_linenumber.getBitVector(2);
    const BitVector & vec_ln4 = bitmap_idx_linenumber.getBitVector(4);

    return ((vec_ln2 | vec_ln4) & (vec_ship | vec_truck)).count();
}

/** 
 * Query 4: Execute the following TPCH query:
 *    SELECT sum(extendedprice * (1.0-discount/100.0))
 *    FROM   lineitem
 *    WHERE  returnflag = 'R' AND
 *           shipmode IN ('REG AIR', 'AIR') AND
 *           quantity BETWEEN 5 AND 10;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @param bitmap_idx_returnflag bitmap index on return flag
 * @param bitmap_idx_shipmode bitmap index on shipmode
 * @param bitmap_idx_quantity bitmap index on quantity
 * 
 * @return returns the sum value.
 */
double
query4_with_index(uint32_t numrows, lineitem_t  *inputtable,
                  BitmapIndex<char>     & bitmap_idx_returnflag,
                  BitmapIndex<char[11]> & bitmap_idx_shipmode,
                  BitmapIndex<int64_t>  & bitmap_idx_quantity)
{
    char regair[11] = "REG AIR";
    BitVector vec_regair = bitmap_idx_shipmode.getBitVector(regair);
    char air[11] = "AIR";
    const BitVector & vec_air = bitmap_idx_shipmode.getBitVector(air);
    const BitVector & vec_retflag = bitmap_idx_returnflag.getBitVector('R');
    
    BitVector vec_selected(numrows);
    for(int64_t q=5; q<=10; q++) {
        vec_selected |= bitmap_idx_quantity.getBitVector(q);
    }
    vec_regair |= vec_air;
    vec_selected &= vec_regair;
    vec_selected &= vec_retflag;

    vec_selected.begin();
    double sum = 0.0;
    register uint32_t idx = 0;
    while(vec_selected.hasNext()){
        if(vec_selected.getNext()) {
            sum += inputtable[idx].extendedprice * (1.0 - inputtable[idx].discount / 100.0);
        }
        idx++;
    }

    return sum;
}
