/**
 * Architektur & Implementierung von DBMS, Aufgabenblatt 7
 *
 * Bitmap Indexing, Query Processing and WAH Compression
 *
 * File        : rawDB.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <sys/time.h>
#include <time.h>
#include <stddef.h>

#include "schema.h"
#include "load.h"
#include "queries.h"
#include "BitmapIndex.h"
#include "BitVector.h"

#define NUM_REPEAT 5

/** Do bitmap indexing for the given indexname and measure time ****/
#define INDEX(IDXNAME)                                              \
    do {                                                            \
        gettimeofday(&start, NULL);                                 \
        nWords = IDXNAME.buildIndex(inputtable);                    \
        gettimeofday(&end, NULL);                                   \
        diff_usec = (end.tv_sec*1000000L + end.tv_usec)             \
                    - (start.tv_sec*1000000L+start.tv_usec);        \
        printf("==> indexing time for %s in microseconds = %lf \n", \
               #IDXNAME, (double)(diff_usec));                      \
        printf("==> memory usage for index %s is %d words \n",      \
               #IDXNAME, nWords);                                   \
    } while(0)

/** Run given query number of times and measure avg execution time */
#define QUERY(QUERYNAME, ...)                                       \
    do {                                                            \
        repeat = NUM_REPEAT;                                        \
        gettimeofday(&start, NULL);                                 \
        while(repeat--){                                            \
            retVal = QUERYNAME(numrows, inputtable, ##__VA_ARGS__); \
        }                                                           \
        gettimeofday(&end, NULL);                                   \
        printf("==> %s return value = %lf \n", #QUERYNAME, retVal); \
        diff_usec = (end.tv_sec*1000000L + end.tv_usec)             \
                    - (start.tv_sec*1000000L+start.tv_usec);        \
        printf("==> %s execution time in microseconds = %lf \n",    \
               #QUERYNAME, (double)(diff_usec)/NUM_REPEAT);         \
    } while(0)


int 
main(int argc, char **argv)
{
   uint32_t     numrows, nWords; 
   lineitem_t * inputtable;
   double       retVal = 0;
   int32_t      repeat = NUM_REPEAT;

   /* Timing related variables */
   struct  timeval start, end;
   int64_t diff_usec;

   /********** load input data **********/
   printf("==> loading data file \"%s\"...\n", DATA_FILE); 
   gettimeofday(&start, NULL); 
   if (load_lineitems_rows((char*)DATA_FILE, &inputtable, &numrows)) {
      fprintf(stderr, "==> error while reading data file: \"%s\"\n", DATA_FILE);
      exit(1); 
   }
   gettimeofday(&end, NULL); 
   printf("read %u rows in %d seconds\n", numrows, 
          (int)(end.tv_sec-start.tv_sec));
   /********** loaded input data **********/

   /********** Construct indexes to be used later **********/
#if defined(WAH) && WAH == 1
   printf("***** Bitmap indexes will be built using WAH compression *****\n");
#else
   printf("***** Bitmap indexes will be built without compression *****\n");
#endif

   // 1) bitmap index on shipmode which is a fixed text of 10 chars
   FieldType<char[11]> field_shipmode(10, offsetof(lineitem_t, shipmode));
   BitmapIndex<char[11]> bitmap_idx_shipmode(field_shipmode, numrows);

   // 2) bitmap index on linenumber which is an int [1-7]
   FieldType<int32_t> field_linenumber(8, offsetof(lineitem_t, linenumber));
   BitmapIndex<int32_t> bitmap_idx_linenumber(field_linenumber, numrows);

   // 3) bitmap index on quantity which is an int [1-50]
   FieldType<int64_t> field_quantity(51, offsetof(lineitem_t, quantity));
   BitmapIndex<int64_t> bitmap_idx_quantity(field_quantity, numrows);

   // 4) bitmap index on returnflag which is a char {'R', 'N', 'A'}
   FieldType<char> field_returnflag(3, offsetof(lineitem_t, returnflag));
   BitmapIndex<char> bitmap_idx_returnflag(field_returnflag, numrows);
   
   // 5) test for high cardinality attribute
   // todo


   /********** Now actually build the indexes **********/
   // create bitmap index on linenumber
   INDEX(bitmap_idx_linenumber);

   // create bitmap index on shipmode
   INDEX(bitmap_idx_shipmode);

   // create bitmap index on quantity
   INDEX(bitmap_idx_quantity);

   // create bitmap index on returnflag
   INDEX(bitmap_idx_returnflag);

   /********** Execute the queries without indexes **********/
   // execute query1 without index
   QUERY(query1);

   // execute query2 without index
   QUERY(query2);

   // execute query3 without index
   QUERY(query3);

   // execute query4 without index
   QUERY(query4);

   /********** Execute the queries with bitmap indexes **********/
   // execute query1 with bitmap index on linenumber
   QUERY(query1_with_index,
         bitmap_idx_linenumber);

   // execute query2 with bitmap index on linenumber, shipmode & quantity
   QUERY(query2_with_index,
         bitmap_idx_linenumber,
         bitmap_idx_shipmode,
         bitmap_idx_quantity);

   // execute query3 with bitmap index on linenumber & shipmode
   QUERY(query3_with_index,
         bitmap_idx_linenumber,
         bitmap_idx_shipmode);

   // execute query4 with bitmap index on returnflag, shipmode & quantity
   QUERY(query4_with_index,
         bitmap_idx_returnflag,
         bitmap_idx_shipmode,
         bitmap_idx_quantity);

   /********** cleanup before exit **********/
   free(inputtable);

   return 0;
}
